import json
import os
from rich.console import Console
from rich.panel import Panel
from ftio.api.gekkoFs.jit.jitsettings import JitSettings


class JitTime:
    def __init__(self) -> None:
        self._app = 0
        self._stage_in = 0
        self._stage_out = 0

    @property
    def app(self):
        return self._app

    # setting the apps
    @app.setter
    def app(self, app):
        self._app = app

    # deleting the values
    @app.deleter
    def app(self):
        del self._app

    @property
    def stage_out(self):
        return self._stage_out

    # setting the stage_outs
    @stage_out.setter
    def stage_out(self, stage_out):
        self._stage_out = stage_out

    # deleting the stage_outs
    @stage_out.deleter
    def stage_out(self):
        del self._stage_out

    @property
    def stage_in(self):
        return self._stage_in

    # setting the stage_ins
    @stage_in.setter
    def stage_in(self, stage_in):
        self._stage_in = stage_in

    # deleting the stage_ins
    @stage_in.deleter
    def stage_in(self):
        del self._stage_in

    def print_time(self):
        console = Console()
        text = (
            f"App time      : {self._app}s\n"
            f"Stage out time: {self._stage_out}s\n"
            f"Stage in time : {self._stage_in}s\n"
            "--------------------------------\n"
            f"Total time : {self._app + self._stage_out + self._stage_in}s\n"
        )
        console.print(
            Panel.fit(
                "[cyan]" + text,
                title="Total Time",
                style="white",
                border_style="white",
                title_align="left",
            )
        )
        return text

    def to_dict(self):
        return {
            "app": self._app,
            "stage_in": self._stage_in,
            "stage_out": self._stage_out,
        }

    def dump_json(self, settings: JitSettings):
        data = {**self.to_dict(), **settings.to_dict()}
        parent = os.path.dirname(settings.log_dir)
        json_path = os.path.join(parent, "result.json")
        try:
            with open(json_path, "r+") as file:
                try:
                    existing_data = json.load(file)
                except json.JSONDecodeError:
                    existing_data = []

                # Check if there is an existing entry with the same number of nodes and mode
                for entry in existing_data:
                    if entry.get("nodes") == data["nodes"]:
                        for i, data_entry in enumerate(entry["data"]):
                            if data_entry["mode"] == data["mode"]:
                                entry["data"][i] = data  # Overwrite the existing data
                                break
                        else:
                            entry["data"].append(data)  # Add new entry if no mode match
                        break
                else:
                    # If no entry with the same nodes exists, add a new one
                    existing_data.append({"nodes": data["nodes"], "data": [data]})

                # Write the updated data back to the file
                file.seek(0)
                json.dump(existing_data, file, indent=4)
                file.truncate()

        except FileNotFoundError:
            # If the file does not exist, create it with the new data
            with open(json_path, "w") as file:
                json.dump([{"nodes": data["nodes"], "data": [data]}], file, indent=4)


    def print_and_save_time(self, settings: JitSettings):
        # get the time
        text = self.print_time()
        # write it out to the file
        time_log_file = os.path.join(settings.log_dir, "time.log")
        with open(time_log_file, "a") as file:
            file.write(text)

        self.dump_json(settings)



import csv

# Function to read a CSV file and store its columns in separate arrays
def read_csv_file(file_path):
    arrays = {}
    
    with open(file_path, 'r') as csvfile:
        reader = csv.DictReader(csvfile)
        
        # Initialize arrays based on headers
        for header in reader.fieldnames:
            arrays[header] = []
        
        # Populate arrays with data from the CSV file
        for row in reader:
            for header in reader.fieldnames:
                arrays[header].append(row[header])
    
    return arrays

# # Example usage
# file_path = 'data.csv'
# arrays = read_csv_file(file_path)

# # Print the arrays
# for key, array in arrays.items():
#     print(f"{key}: {array}")



import os
import sys
import time
import json
from multiprocessing import Pool, cpu_count
from concurrent.futures import ProcessPoolExecutor, as_completed
import pandas as pd
from rich.console import Console
from rich.progress import (
    Progress,
    SpinnerColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TaskProgressColumn,
    TextColumn,
    BarColumn,
)
from ftio.api.trace_analysis.trace_ftio_v2 import main as trace_ftio
from ftio.api.trace_analysis.trace_analysis import (
    convert_dict,
    flatten_dict,
    statistics,
)

# Initialize the console for printing
console = Console()


def process_file(file_path, argv, settings, index=0):
    error = ""
    try:
        # Call your trace_ftio function (adjust the import and call as necessary)
        res = trace_ftio([file_path] + argv, settings["verbose"], settings["json"])

        # Create the new file name by replacing the pattern
        base_name = os.path.basename(file_path)
        
        # if input file is not a json, save ftio result
        if not settings["json"] and settings["save"]:
            json_file = base_name.replace(f"_signal_{settings["name"]}.csv", f"_freq_{settings["name"]}.json")
            json_path = os.path.join(os.path.dirname(file_path), json_file)
            # Convert NumPy arrays to lists and save the ftio results
            data_converted = convert_dict(res)
            if json_file.endswith("json"):
                with open(json_path, "w") as file:
                    json.dump(data_converted, file, indent=4)
            else:
                console.print(f"[bold red]Cannot dump Json file in {json_path}[/]")

        flat_res = flatten_dict(res)
        try:
            if settings["json"]:
                flat_res["job_id"] = base_name.removesuffix(".json")
            else:
                flat_res["job_id"] = base_name.split("_")[0]
            flat_res["file"] = file_path
        except:
            flat_res["job_id"] = "??"
            flat_res["file"] = "??"
            console.print("[bold red]Unable to extract job id[/]")
    except Exception as e:
        console.print(f"\n[bold red]Error processing file {file_path}: {e}[/]")
        error = str(e)
        flat_res = None

    return flat_res, index, file_path, error


def main(argv=sys.argv[1:]) -> None:
    settings={
        "verbose": False,
        "json": False,
        "save": False,
        "name": "plafrim",
        "num_procs": -1,
        "res_path": ".",
        "freq": 10,
        "folder_path":""
    }

    # Specify the name with -n
    if "-p" in argv:
        index = argv.index("-p")
        settings["num_procs"] = int(argv[index + 1])
        argv.pop(index)
        argv.pop(index)
    if "-n" in argv:
        index = argv.index("-n")
        settings["name"] = str(argv[index + 1])
        argv.pop(index)
        argv.pop(index)
    if "-v" in argv:
        index = argv.index("-v")
        argv.pop(index)
        settings["verbose"] = True
    if "-j" in argv:
        index = argv.index("-j")
        argv.pop(index)
        settings["json"] = True
    if "-s" in argv:
        index = argv.index("-s")
        argv.pop(index)
        settings["save"] = True
    if "-o" in argv:
        index = argv.index("-o")
        settings["res_path"] = str(argv[index + 1])
        argv.pop(index)
        argv.pop(index)
    if "-h" in argv:
        console.print(
            "Usage:  parallel_trace_analysis  <dir>\n\n"
            "-n <str>: filter according if they contain the name\n"
            "--time-step <float>: specifies the value of the implicit time steps between the samples\n"
            "-o <str>: Output dir location\n"
            "-j <bool>: Enables JSON search\n"
            "-v <bool>: verbose\n"
            "-s <bool>: save the FTIO result for each file in a file that contains the name _freq_\n"
            "All ftio options (see ftio -h)\n\n"
            )
        sys.exit()
    if "-f" in argv: #save the freq
        index = argv.index("-f")
        settings["freq"] = str(argv[index + 1])

    start_time = time.time()
    pattern = f"_signal_{settings["name"]}.csv"
    df = pd.DataFrame()
    if settings["json"]:
        pattern = ".json"

    # Handle command-line arguments properly
    if len(argv) > 0:
        folder_path = argv[0]
        argv.pop(0)
    else:
        folder_path = "/d/github/FTIO/ftio/api/trace_analysis/plafrim"

    folder_path = os.path.abspath(folder_path)
    settings["folder_path"] = folder_path # log this
    console.print(f"[bold green]Path is: {folder_path}[/]")
    # Create a list to store all matching csv files
    trace_files = []
    # Walk through all subdirectories and find matching files
    for dirpath, _, filenames in os.walk(folder_path):
        for filename in filenames:
            if filename.endswith(pattern) and "_freq_" not in filename:
                trace_files.append(os.path.join(dirpath, filename))

    if not trace_files:
        console.print(f"[bold red]No files matched the pattern: {pattern}![/]")
        return

    total_files = len(trace_files)
    # Create a progress bar
    progress = Progress(
        SpinnerColumn(),
        TextColumn(
            "[progress.description]{task.description} ({task.completed}/{task.total})"
        ),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(),
        "[yellow]-- runtime",
        TimeElapsedColumn(),
    )

    try:
        with progress:
            # Use multiprocessing Pool
            if settings["num_procs"] == -1:
                settings["num_procs"] = int(cpu_count() / 2)

            console.print(f"[bold green]Using {settings["num_procs"]} processes[/]\n")
            task = progress.add_task("[green]Processing files", total=total_files)

            # List to store failed file details
            failed_files = []

            # use future or apply_async
            future = (
                True  # managing tasks without blocking others due to slow processes
            )

            if future:
                counter = 0
                with ProcessPoolExecutor(max_workers=settings["num_procs"]) as executor:
                    # Submit tasks to the executor
                    futures = {
                        executor.submit(
                            process_file, file_path, argv, settings, i): i
                        for i, file_path in enumerate(trace_files)
                    }

                    # Process results as they complete
                    for future in as_completed(futures):
                        # index = futures[future]
                        try:
                            flat_res, index, file_path, error = future.result(
                                timeout=120
                            )
                        except TimeoutError:
                            index = futures[future]
                            flat_res = None
                            error = f"[red bold] {trace_files[index]} reached timeout"
                        if flat_res:  # Process result
                            df = pd.concat(
                                [df, pd.DataFrame([flat_res])], ignore_index=True
                            )
                        if error:  # Log any failed files
                            failed_files.append((file_path, error))
                        # Update progress
                        progress.console.print(
                            f"Processed ({index + 1}/{len(trace_files)}): {trace_files[index]}"
                        )
                        counter += 1
                        progress.update(task, completed=counter)
            else:
                with Pool(processes=settings["num_procs"]) as pool:
                    # Pass the index and total files to process_file
                    results = [
                        pool.apply_async(
                            process_file, (file_path, argv, settings, index)
                        )
                        for i, file_path in enumerate(trace_files)
                    ]

                    for result in results:
                        flat_res, index, file_path, error = result.get()
                        if flat_res:  # Process result
                            df = pd.concat([df, pd.DataFrame([flat_res])], ignore_index=True)
                        if error:  # Log any failed files
                            failed_files.append((file_path, error))
                        
                        # Update progress
                        progress.console.print(
                            f"Processed ({index + 1}/{total_files}): {trace_files[index]}"
                        )
                        progress.update(task, completed=index + 1)

        # After processing, log the files that failed
        if failed_files:
            progress.console.print(
                f"\n[bold]{total_files} files processed: [/]\n"
                f" - [bold green]{total_files - len(failed_files)} files processed successfully![/]\n"
                f" - [bold red]{len(failed_files)} files failed[/]\n"
            )
            console.print("\n[bold yellow]The following files failed to process:[/]")
            with open(f"{settings["res_path"]}/log.err", "w") as log:
                for file_path, error in failed_files:
                    console.print(f"[bold red]{file_path}[/]")
                    log.write(f"{file_path}: {error}\n")
        else:
            progress.console.print(
                "\n[bold green]All files processed successfully![/]\n"
            )

        console.print(
            f"[blue]FTIO total time:[/] {time.time() - start_time:.4f} seconds\n"
            f"[blue]Location:[/] {folder_path}\n"
            f"[blue]Pattern:[/] {pattern}\n"
        )
        elapsed_time = f"Execution time {time.time() - start_time:.4f} seconds"
        statistics(df, elapsed_time, settings)
    except KeyboardInterrupt:
        progress.console.print("[bold red]Keyboard interrupt![/]\n")
        statistics(df, "", settings)
        sys.exit()
    console.print(f"[blue]Execution time:[/] {time.time() - start_time:.4f} seconds")


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import numpy as np
import json
import re
from time import process_time
from ftio.freq.helper import MyConsole
from ftio.api.metric_proxy.req import MetricProxy


CONSOLE = MyConsole()
CONSOLE.set(True)

def parse(file_path, match="proxy_component_critical_temperature_celcius")-> tuple[np.ndarray,np.ndarray]:
    b_out = np.array([])
    t_out = np.array([])
    try:
        with open(file_path, 'r') as json_file:
            json_data = json.load(json_file)
    except FileNotFoundError:
        print(f"Error: File '{file_path}' not found.")
        return b_out,t_out
    except json.JSONDecodeError:
        print(f"Error: Unable to decode JSON from file '{file_path}'. Check if the file is valid JSON.")
        return b_out,t_out

    b_out,t_out = extract(json_data, match)

    if len(b_out) == 0:
        print("No match found. Exciting\n")
        exit(0)
    return b_out,t_out


def extract(json_data, match, verbose=False):
    b_out = np.array([])
    t_out  = np.array([])
    for key, value in json_data.items():
        if isinstance(value, dict):
            b_out,t_out = extract(value, match, verbose)
            if len(b_out) > 0:
                break
        else:
            if match == key:
                if verbose:
                    print(f"matched {key}")
                x = np.array(value)
                x = np.nan_to_num(x=x, copy=False, nan=0.0, posinf=0.0, neginf=0.0)
                t_out = x[:,0]
                b_out = x[:,1]
                #remove None from b
                b_out[b_out==None]= 0
                #reduce to derivative
                if "deriv" not in key:
                    if verbose:
                        print("removing aggregation")
                    b_shifted = b_out[:-1]
                    b_shifted = np.insert(b_shifted,0,0)
                    b_out =  b_out - b_shifted
                    break
    return b_out,t_out


def filter_metrics(json_data,filter_deriv:bool=True, exclude=None, scale_t:float = 1, rename:dict = {}):
    out = {}
    t = process_time()
    metrics = json_data['metrics'].keys()
    # extract either derive or all but not all
    if filter_deriv:
        metrics = clean_metrics(metrics)

    if exclude:
        old_length = len(metrics)
        metrics = [metric for metric in metrics if all(n not in metric for n in exclude)]
        text=', '.join([str(item) for item in exclude])
        CONSOLE.info(f"[blue]\nExcluded matches for: \\[{text}]\nMetrics reduced further from {old_length} to {len(metrics)}[/]")
    
    for metric in metrics:
        b_out,t_out = extract(json_data,metric, False)
        out[metric]=[b_out,t_out*scale_t]

    # rename keys if only one metric passed
    if exclude:
        if "func" not in exclude:
            keys_to_rename = []
            for metric in out:
                if "func" in metric:
                    keys_to_rename.append((metric, "f_"+re.findall(r'func__(.*?)__', metric)[0]))
            for old_key, new_key in keys_to_rename:
                original_new_key = new_key  
                suffix = 1
                while new_key in out:
                    new_key = f"{original_new_key}_{suffix}"
                    suffix += 1
                out[new_key] = out.pop(old_key)

        if  len(set(["time", "hits", "size"]) & set(exclude)) == 2:
            keys_to_rename = [(metric, metric.rsplit("__", 1)[-1]) for metric in out]
            # Perform renaming after collecting all keys
            if keys_to_rename:
                CONSOLE.info(f"[blue]\nRenaming Metrics: Removing [{keys_to_rename[-1][-1]}] from names[/]")

            for old_key, new_key in keys_to_rename:
                out[new_key] = out.pop(old_key)

    elapsed_time = process_time() - t
    CONSOLE.info(f"[blue]Parsing time: {elapsed_time} s[/]")
    
    return out


def parse_all(file_path:str,filter_deriv:bool=True, exclude=None, scale_t:float = 1)-> dict:
    """parses all metrics from proxy

    Args:
        file_path (str): pass to proxy JSON file
        filter_deriv (bool, optional): Removes the metrics in case a similar metrics, which start with deriv is presented. Defaults to True.
        exclude (list,optional): list of metrics to exclude
        scale_t (float, optional): scale time unit (default 1). Default unit is "s"

    Returns:
        dict: parsed metrics with 2D numpy array
    """
    CONSOLE.info(f"\n[blue]Current file: {file_path}[/]")
    if scale_t != 1:
        CONSOLE.info(f"\n[yellow]Scaling time by: {scale_t}[/]")
    try:
        with open(file_path, 'r') as json_file:
            json_data = json.load(json_file)
    except FileNotFoundError:
        print(f"Error: File '{file_path}' not found.")
        return {}
    except json.JSONDecodeError:
        print(f"Error: Unable to decode JSON from file '{file_path}'. Check if the file is valid JSON.")
        return {}

    return filter_metrics(json_data,filter_deriv,exclude, scale_t)


def load_proxy_trace_stdin(deriv_and_not_deriv:bool=True, exclude=None):
    try:
    # Read JSON from stdin
        data = json.load(sys.stdin)
    except json.JSONDecodeError as e:
        print(f"Error parsing JSON: {e}")
    else:
        return filter_metrics(data,deriv_and_not_deriv,exclude)


def clean_metrics(metrics:str):
    deriv_metrics =  [metric for metric in metrics if metric.startswith("deriv")]
    non_deriv_metrics = [metric for metric in metrics if not metric.startswith("deriv")]
    cleaned_metrics = [metric for metric in metrics if not (metric in non_deriv_metrics and "deriv__" + metric in deriv_metrics)]
    CONSOLE.info(f"[blue]Metrics reduced from {len(metrics)} to {len(cleaned_metrics)}[/]")
    return cleaned_metrics



def get_all_metrics(job_id):
    mp = MetricProxy()
    metrics = {}
    all_metrics= mp.metric(job_id)
    for metric in all_metrics:
        value = mp.trace_metric(job_id,metric)
        t_out = np.array([item[0] for item in value])
        b_out = np.array([item[1] for item in value])
        # derive b
        b_out = numerical_derivative(t_out,b_out)
        metrics[metric]=[b_out,t_out]
        
    return metrics

# Calculate the numerical derivative using central differences
def numerical_derivative(t, f):
    n = len(t)
    df_dt = np.zeros(n)
    if n > 10:
        # Forward difference for the first point
        df_dt[0] = (f[1] - f[0]) / (t[1] - t[0])

        # Central differences for the interior points
        for i in range(1, n - 1):
            df_dt[i] = (f[i + 1] - f[i - 1]) / (t[i + 1] - t[i - 1])

        # Backward difference for the last point
        df_dt[-1] = (f[-1] - f[-2]) / (t[-1] - t[-2])  
    return df_dt



import argparse
import numpy as np
from multiprocessing import Manager, cpu_count
from concurrent.futures import ProcessPoolExecutor, as_completed

# from rich.progress import Progress
from ftio.api.metric_proxy.parse_proxy import parse_all,get_all_metrics, filter_metrics
from ftio.api.metric_proxy.helper import extract_data, create_process_bar
from ftio.api.metric_proxy.proxy_analysis import phases, phases_and_timeseries
from ftio.api.metric_proxy.plot_proxy import (
    heatmap,
    scatter2D,
    scatter,
    heatmap_2,
    density_heatmap,
    plot_timeseries_metrics,
)
from ftio.api.metric_proxy.proxy_cluster import optics,dbscan
from ftio.api.metric_proxy.req import MetricProxy
from ftio.prediction.tasks import ftio_task, ftio_task_save
from ftio.prediction.helper import print_data
from ftio.freq.helper import MyConsole


CONSOLE = MyConsole()
CONSOLE.set(True)


def main(args:argparse.Namespace) -> None:
    argv=args.argv 
    # shows the results from FTIO
    show = False
    # pools or future
    pools = False
    
    # ---------------------------------
    # Modification area
    # ---------------------------------
    argv.extend(["-e","no"])
    # argv = ["-e", "mat"]  # ["-e", "plotly"]
    # # finds up to n frequencies. Comment this out to go back to the default version
    # argv.extend(["-n", "10"])
    # ---------------------------------

    if args.proxy:
        mp = MetricProxy()
        jobs = mp.jobs()
        job_id = jobs[0]["jobid"]
        jsonl = mp.trace(job_id)
        metrics = filter_metrics(jsonl,filter_deriv=True)

        # Workaround: proxy needs to be running
        # metrics = get_all_metrics('4195024897')
    else:
        metrics = parse_all(args.file , filter_deriv=True,exclude=["size","hits","proxy"], scale_t=1e-3) # "mpi"

    CONSOLE.print("[blue]\nSettings:\n---------[/]")
    CONSOLE.print(f"[blue] - parallel: {not args.disable_parallel}[/]")
    CONSOLE.print(f"[blue] - future: {not pools}[/]")
    CONSOLE.print(f"[blue] - proxy: {args.proxy}[/]\n")
    ranks = 32

    # plot the metrics if needed
    # _ = plot_timeseries_metrics(metrics, 2000,500)

    if not args.disable_parallel:
        data = execute_parallel(metrics, argv, ranks, show, pools)
    else:
        data = execute(metrics, argv, ranks, show)
    
    post(data, metrics, argv)


def execute_parallel(
    metrics: dict, argv: list, ranks: int, show: bool = False, pools: bool = False
)  :
    # parallel
    manager = Manager()
    data = manager.list()
    counter = 0
    total_metrics = len(metrics)
    progress = create_process_bar(total_metrics)
    task = progress.add_task("[green]Processing metrics", total=total_metrics)
    with progress:
        try:
            if pools:
                # with Progress() as progress:
                #     task = progress.add_task("[cyan]Metrics handled", total=len(metrics.keys()))
                for metric, arrays in metrics.items():
                    # with ProcessPoolExecutor(max_workers=80) as executor:
                    with ProcessPoolExecutor() as executor:
                        _ = executor.submit(
                            ftio_task_save, data, metric, arrays, argv, ranks, show
                        )
                        # progress.update(task, advance=1)
                        counter += 1
                        progress.update(task, completed=counter)
            else: #use futures
                with ProcessPoolExecutor(max_workers=cpu_count()-2) as executor:
                    futures = {
                            executor.submit(
                                ftio_task_save, data, metric, arrays, argv, ranks, show): metric 
                            for metric, arrays in metrics.items()
                        }
                    for future in as_completed(futures):
                        counter += 1
                        progress.update(task, completed=counter)
        except KeyboardInterrupt:
            print_data(data)
            print("-- done -- ")
            exit()

    return data


def execute(metrics: dict, argv: list, ranks: int, show: bool):
    data = []
    save = True
    check = True
    error_counter = 0
    counter = 0
    total_files = len(metrics)
    progress = create_process_bar(total_files)
    
    with progress:
        task = progress.add_task("[green]Processing metrics", total=total_files)
        for metric, arrays in metrics.items():
            if check:
                decreasing_order = np.all(arrays[1][-1] >= arrays[1][1])
                # negative = np.all(arrays[0] <= 0)
                if not decreasing_order: #or not  negative:
                    error_counter += 1
                    # err = "[bold red] Negative metric" if not negative else "[bold yellow] time not decreasing"
                    err = "[bold yellow] time not decreasing"
                    CONSOLE.print(f"[bold red]- {error_counter}. Error in {metric}:{err}[/]")
                    continue
            if save:
                # save stuff in queue, data is non empty
                ftio_task_save(data, metric, arrays, argv, ranks, show)
            else:
                # skip saving
                ftio_task(metric, arrays, argv, ranks, show)
            counter += 1
            progress.update(task, completed=counter)

    return data 

def post(data, metrics, argv):
    if data:
        # print_data(data)
        phases(data, argv)
        phases_and_timeseries(metrics, data, argv)
        df = extract_data(data)
        heatmap(data)
        scatter2D(df)
        scatter(
            df, x="Phi", y="Dominant Frequency", color="Confidence", symbol="Metric"
        )
        _ = optics(df,"Phi","Dominant Frequency")
        _ = dbscan(df,"Phi","Dominant Frequency",0.1)
        density_heatmap(data)
        heatmap_2(data)
    else:
        CONSOLE.print("[bold red] No data[/]")

if __name__ == "__main__":
    
    # file = "/d/github/FTIO/ftio/api/metric_proxy/traces/alberto_unito/bench_8x144.json"
    # file = "/d/github/FTIO/ftio/api/metric_proxy/traces/jb_traces/WACOM_PROCESS_BASED_json/wacommplusplus.36procs.trace.json"
    # file = "/d/sim/metric_proxy/traces/Mixed_1x8_5.json"file = /d/github/FTIO/ftio/api/metric_proxy/new_traces/imbio.json"
    # file = "/d/github/FTIO/ftio/api/metric_proxy/new_traces/imbio.json"
    # file = "/d/github/FTIO/ftio/api/metric_proxy/new_traces/wacom.json"
    file = "/d/github/FTIO/ftio/api/metric_proxy/new_traces/wacoml.json"
    # file = "/d/github/FTIO/ftio/api/metric_proxy/new_traces/lulesh_8_procs.json"
    # file = "/d/github/FTIO/ftio/api/metric_proxy/new_traces/lulesh_27_procs.json"
    

    parser = argparse.ArgumentParser(description="Executes FTIO in parallel on a JSON file from the proxy or by directly communicating with the proxy")
    parser.add_argument(
        '--file',
        type=str,
        nargs="?",  # '*' allows zero or more filenames
        default=file,
        help="The paths to the JSON file from the proxy"
    )
    parser.add_argument('--proxy', action = 'store_true', default=False, help="parallel or not")
    parser.add_argument('--disable_parallel', action = 'store_true', default=False, help="parallel or not")
    args, unknown = parser.parse_known_args()
    args.argv = unknown
    main(args)
    


import numpy as np
import pandas as pd
import plotly.express as px
from plotly.subplots import make_subplots
import plotly.graph_objects as go
from ftio.plot.helper import format_plot


def plot_spectrum(
    amp: np.ndarray, freq: np.ndarray, mode: str = "Amplitude", percent: bool = False
):
    template = "plotly"
    name = "Amplitude" if "amp" in mode.lower() else "Power"
    unit = ""
    if percent:
        name = "Normed " + name
        unit = unit + " (%)"
        if amp.sum() > 0:
            amp = 100 * amp / amp.sum()

    df_tmp = pd.DataFrame(
        data={
            "A": amp,
            "freq": freq,
        }
    )
    # fig_tmp = px.bar(df_tmp, x="freq", y="A", color="A", color_continuous_scale='Bluered')
    # fig_tmp = px.bar(df_tmp, x="freq", y="A", color="A", color_continuous_scale=["rgb(0,10,130)", "rgb(130,30,130)", "rgb(255,50,50)"])
    # fig_tmp = px.bar(df_tmp, x="freq", y="A", color="A", color_continuous_scale="plasma")
    # cyan - purple
    # fig_tmp = px.bar(df_tmp, x="freq", y="A", color="A", color_continuous_scale=[ "rgb(90, 40, 90)", "rgb(113, 221, 255)" ])
    fig_tmp = px.bar(
        df_tmp,
        x="freq",
        y="A",
        color="A",
        color_continuous_scale=["rgb(0,50,150)", "rgb(150,50,150)", "rgb(255,50,0)"],
    )

    fig_tmp.update_traces(
        marker_line=dict(width=0.2, color="black"),
        hovertemplate="<b>Frequency</b>: %{x:.2e} Hz <br><b>"
        + f"{name}"
        + "</b>: %{y:.2e}"
        + f"{unit}",
    )
    fig_tmp.update_layout(
        xaxis_title="Frequency (Hz)",
        yaxis_title=name + unit,
        coloraxis_colorbar=dict(yanchor="top", y=1, x=0, ticks="outside"),
        template=template,
    )
    fig_tmp = format_plot(fig_tmp)
    return fig_tmp, name + unit


def plot_both_spectrums(args, freq: np.ndarray, amp: np.ndarray, full: bool = True):
    fig_1 = make_subplots(
        rows=2,
        cols=2,
        specs=[[{"colspan": 2}, None], [{"colspan": 2}, None]],
    )

    start = 1

    if not full:
        indecies = np.arange(start, int(len(amp) / 2) + 1)
        freq = freq[indecies]
        amp = 2 * amp[indecies]
        amp[-1] = amp[-1] / 2
        if start == 0:
            amp[0] = amp[0] / 2

    if args.psd:
        power = amp
        amplitude = np.sqrt(amp) * len(amp)
    else:
        power = amp * amp / len(amp)
        amplitude = amp

    fig_tmp, name_plt = plot_spectrum(amplitude, freq, "Amplitude", False)
    layout = fig_tmp.layout
    for trace in list(fig_tmp.select_traces()):
        trace.update(marker={"coloraxis": "coloraxis"})
        fig_1.append_trace(trace, row=1, col=1)
    fig_tmp, name_plt1 = plot_spectrum(power, freq, "Power", True)
    for trace in list(fig_tmp.select_traces()):
        trace.update(marker={"coloraxis": "coloraxis2"})
        fig_1.append_trace(trace, row=2, col=1)
    fig_1.update_layout(layout)
    fig_1.update_layout(
        coloraxis={
            "colorbar": {
                "x": 1,
                "len": 0.4,
                "y": 0.8,
            },
        },
        coloraxis2={
            "colorbar": {
                "x": 1,
                "len": 0.4,
                "y": 0.2,
            },
        },
    )

    return fig_1, [name_plt, name_plt1]


def plot_one_spectrum(
    psd_flag: bool,
    freq: np.ndarray,
    amp: np.ndarray,
    full: bool = True,
    percent: bool = False,
) -> go.Figure:
    if not full:
        n_samples = len(freq)
        last = int(n_samples / 2) - 1
        amp = 2 * amp[0 : last + 1]
        amp[last] = amp[last] / 2
        amp[0] = amp[0] / 2
        freq = freq[0 : int(n_samples / 2)]

    if psd_flag:
        amp = amp * amp / len(amp)
        mode = "Power"
    else:
        mode = "Amplitude"

    fig_tmp, _ = plot_spectrum(amp, freq, mode, percent)
    return fig_tmp



import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from ftio.freq.helper import MyConsole
from ftio.plot.print_html import PrintHtml
from ftio.plot.helper import format_plot_and_ticks
from ftio.api.metric_proxy.phasemode import PhaseMode


CONSOLE = MyConsole()
CONSOLE.set(True)


def phases_and_timeseries(metrics, data, argv=[]):
    phasemode_list, t = classify_waves(data, True)

    if argv and '-n' in argv:
        n = argv[argv.index('-n') + 1]
        plot_waves_and_timeseries(argv,metrics, phasemode_list, t, n)
    else:
        plot_waves_and_timeseries(argv,metrics, phasemode_list, t)


def phases(data, argv=[]):
    phasemode_list, t = classify_waves(data)

    if argv and '-n' in argv:
        n = argv[argv.index('-n') + 1]
        plot_waves(argv,phasemode_list, t, n)
    else:
        plot_waves(argv,phasemode_list, t)


def classify_waves(data, normed=True):
    phasemode_list = []
    phasemode_list.append(
        PhaseMode(
            'Network',
            [
                'network',
                'mpi_bcast',
                'sock',
                'send',
                'listen',
                'recieve',
                'gather',
                'reduce',
                'cast',
            ],
        )
    )
    phasemode_list.append(PhaseMode('write', ['write']))
    phasemode_list.append(PhaseMode('read', ['read']))
    phasemode_list.append(
        PhaseMode(
            'I/O',
            [
                'io',
                'disk',
                'stat',
                'cntl',
                'read',
                'write',
                'openat',
                'mmap',
                'seek',
                'close',
            ],
        )
    )
    phasemode_list.append(PhaseMode('Temperature', ['temp']))
    phasemode_list.append(PhaseMode('CPU', ['cpu']))
    phasemode_list.append(PhaseMode('Wait', ['poll', 'barrier']))
    seen = []
    for mode in phasemode_list:
        seen.extend(mode.matches)
    other = [x['metric'] for x in data if all(y not in x['metric'] for y in seen)] 
    phasemode_list.append(PhaseMode('Other', other))

    sampling_freq = np.NaN
    t_s = np.inf
    t_e = 0
    for d in data:
        if len(d['dominant_freq']) > 0 and len(d['conf']) > 0:
            if np.isnan(sampling_freq):
                sampling_freq = d['freq']
                t_s = min(d['t_start'],t_s)
                t_e = max(d['t_end'],t_e)

            add_metric(phasemode_list, d)

    if not np.isnan(sampling_freq):
        t = np.arange(t_s, t_e, 1 / sampling_freq)
        add_time(phasemode_list, t)
    else:
        CONSOLE.print('[red] No metrics classified. Try increasing the sampling frequency[/]')    
        exit(0)

    text = print_len(phasemode_list, data)
    CONSOLE.print(f'[blue]{text}[/]')
    aggregate_wave_for_all_modes(phasemode_list, t, normed)

    return phasemode_list, t


def plot_waves(argv:list, arr: list[PhaseMode], t, n=None):
    plot_mode = False
    if argv and '-e' in argv:
        plot_mode = "mat" in argv[argv.index('-e') + 1]

    if plot_mode:
        # Create a figure and axis using variables
        fig, ax = plt.subplots()

        # Loop through each mode in arr and plot
        for mode in arr:
            ax.plot(t, mode.wave, marker='o', label=mode.name)  # Plot with markers


        # Set axis titles
        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Normed metrics')

        # Set the legend title if n is provided
        if n:
            ax.legend(title=f'{n} Frequencies')
        else:
            ax.legend()

        plt.show()
    else:
        fig = go.Figure()
        for mode in arr:
            fig.add_trace(
                go.Scatter(x=t, y=mode.wave, mode='lines+markers', name=mode.name)
            )

        fig.update_layout(
            xaxis_title='Time (s)',
            yaxis_title='Normed metrics',
        )
        if n:
            fig.update_layout(legend_title_text=f'{n} Frequencies')
        fig = format_plot_and_ticks(fig)
        fig.show()
        # fig.write_image("waves.png")


def plot_waves_and_timeseries(argv:list,  metrics: dict, arr: list[PhaseMode], t, n=None):
    names = [] #get_names(arr)

    out = PrintHtml('./', names, outdir='phase_plots')
    out.generate_html_start()
    # t = []
    for mode in arr:
        f = plot_mode(mode,metrics,t,n,True)
        # for fig in f:
        #     fig.show
        out.generate_html_core(mode.name+'.html', f)

    out.generate_html_end()


def plot_mode(mode,metrics,t,n,subfig=False)-> list[go.Figure]:

    f = []
    spec=[{},{}]
    if subfig:
        f.append(make_subplots(rows=2,cols=1))
        spec=[{"row":1,"col":1},{"row":2,"col":1}]
    else:
        f.append(go.Figure())

    f[-1].add_trace(
        go.Scatter(x=t, y=mode.wave, mode='lines+markers', name=mode.name),**spec[0])

    if not subfig:
        f.append(go.Figure())

    for metric, arrays in metrics.items():
        if len(arrays[0]) > 1 and any(n in metric for n in mode.matches):
            f[-1].add_trace(
                go.Scatter(
                    x=arrays[1],
                    y=arrays[0],
                    mode='lines+markers',
                    name=metric,
                    hovertemplate='<i>Time </i>: %{x} s'
                    + '<br><b>Metric</b>: %{y}<br>',
                    legendgroup=metric,
                ),**spec[1],
            )
            wave,name = mode.get_wave(metric)
            f[-1].add_trace(
                go.Scatter(
                    x=mode.get("t"),
                    y=wave,
                    mode='lines+markers',
                    name=name,
                    hovertemplate='<i>Time </i>: %{x} s'
                    + '<br><b>Metric</b>: %{y}<br>',
                    legendgroup=metric,
                ),**spec[0],
            )

    for fig in f:
        fig = format_plot_and_ticks(fig, font=False)
        fig.update_layout(
            xaxis_title='Time (s)',
            yaxis_title=f'{mode.name.capitalize()} Metrics',
            width=1400,
            height=800,
            showlegend=True,
        ) 
    if subfig:
        fig.update_xaxes(title_text='Time (s)', **spec[1])
        fig.update_yaxes(title_text=f'{mode.name.capitalize()} Metrics',**spec[1])

        if f.index(fig) == 0:
            if n:
                fig.update_layout(title=f'{mode.name.capitalize()}: Using {n} Frequencies per metric')
            else:
                fig.update_layout(title=f'{mode.name.capitalize()}: Using Dominant Frequencies')
    return f




def add_metric(arr: list[PhaseMode], d):
    for mode in arr:
        if mode.match(d):
            mode.add(d)
            break

def add_time(arr: list[PhaseMode], t):
    for mode in arr:
        mode.set_time(t)


def get_names(arr:list[PhaseMode])-> list[str]:
    names = []
    for mode in arr:
        names.append(mode.name)
    return names 


def aggregate_wave_for_all_modes(arr: list[PhaseMode], t, normed=True):
    for mode in arr:
        mode.aggregates_waves(t, normed)


def print_len(arr: list[PhaseMode], data) -> str:
    text = '\n'
    total_metrics = 0
    for mode in arr:
        text += f'{len(mode.data):3} {mode.name} metrics\n'
        total_metrics += len(mode.data)

    text += f'---------------------\n{total_metrics:3}/{len(data):3} total metrics with dominant frequencies\n'

    return text



import os
import time

# import multiprocessing
import subprocess
from rich.console import Console

from ftio.api.gekkoFs.jit.setup_check import check_setup
from ftio.api.gekkoFs.jit.setup_helper import (
    check,
    check_port,
    elapsed_time,
    flaged_mpiexec_call,
    get_env,
    get_executable_realpath,
    handle_sigint,
    jit_print,
    mpiexec_call,
    relevant_files,
    reset_relevant_files,
    shut_down,
)
from ftio.api.gekkoFs.jit.execute_and_wait import (
    end_of_transfer_online,
    execute_block,
    execute_background,
    execute_background_and_log,
    execute_background_and_wait_line,
    execute_block_and_monitor,
    get_time,
    monitor_log_file,
    wait_for_file,
    wait_for_line,
)
from ftio.api.gekkoFs.jit.setup_init import init_gekko
from ftio.api.gekkoFs.jit.jitsettings import JitSettings
from ftio.api.gekkoFs.jit.jittime import JitTime

console = Console()


#! Start gekko
#!#################
def start_gekko_daemon(settings: JitSettings) -> None:

    if settings.exclude_daemon:
        console.print(
            f"[bold yellow]####### Skipping Gkfs Demon [/][black][{get_time()}][/]"
        )
    else:
        console.print(
            f"[bold green]####### Starting Gkfs Demon [/][black][{get_time()}][/]"
        )
        # Create host file and dirs for geko
        init_gekko(settings)
        debug_flag = ""
        if settings.debug:
            debug_flag = "--export=ALL,GKFS_DAEMON_LOG_LEVEL=trace"

        if settings.cluster:
            if settings.use_mpirun:
                # mpiexec
                call = (
                    f"{settings.gkfs_daemon} "
                    f"-r {settings.gkfs_rootdir} -m {settings.gkfs_mntdir} "
                    f"-H {settings.gkfs_hostfile}  -c --clean-rootdir -l ib0 -P {settings.gkfs_daemon_protocol}"
                )
                call = mpiexec_call(settings, call, settings.app_nodes)
            else:
                call = (
                    f"srun {debug_flag} --jobid={settings.job_id} {settings.app_nodes_command} --disable-status -N {settings.app_nodes} "
                    #   f"--ntasks={settings.app_nodes*settings.procs_daemon} --cpus-per-task={settings.procs_daemon} --ntasks-per-node={settings.procs_daemon} --overcommit --overlap "
                    f"--ntasks={settings.app_nodes} --cpus-per-task={settings.procs_daemon} --ntasks-per-node=1 --overcommit --overlap "
                    f"--oversubscribe --mem=0 {settings.task_set_0} {settings.gkfs_daemon} -r {settings.gkfs_rootdir} -m {settings.gkfs_mntdir} "
                    f"-H {settings.gkfs_hostfile}  -c --clean-rootdir -l ib0 -P {settings.gkfs_daemon_protocol}"
                )
            if not settings.exclude_proxy:
                # Demon call with proxy
                call += " -p ofi+verbs -L ib0"
        else:  # no cluster mode
            call = (
                f"GKFS_DAEMON_LOG_LEVEL=info GKFS_DAEMON_LOG_PATH={settings.gekko_daemon_log} {settings.gkfs_daemon} -r {settings.gkfs_rootdir} -m {settings.gkfs_mntdir} "
                f"-H {settings.gkfs_hostfile}  -c --clean-rootdir -l lo -P ofi+tcp"
            )
            if not settings.exclude_proxy:
                # Demon call with proxy
                call += " --proxy-listen lo --proxy-protocol ofi+tcp"

        jit_print("[cyan]>> Starting Demons[/]", True)
        # p = multiprocessing.Process(target=execute_background, args= (call, settings.gekko_daemon_log, settings.gekko_daemon_err, settings.dry_run))
        # p.start()
        # if settings.verbose:
        #     _ = monitor_log_file(settings.gekko_daemon_err,"Error Demon")
        #     _ = monitor_log_file(settings.gekko_daemon_log,"Demon")
        _ = execute_background_and_log(
            settings,
            call,
            settings.gekko_daemon_log,
            "daemon",
            settings.gekko_daemon_err,
        )
        if settings.verbose:
            _ = monitor_log_file(settings.gekko_daemon_err, "Error Demon")

        wait_for_file(settings.gkfs_hostfile, dry_run=settings.dry_run)
        console.print("\n")


#! Start Proxy
#!#######################
def start_gekko_proxy(settings: JitSettings) -> None:
    if settings.exclude_proxy:
        console.print(
            f"[bold yellow]####### Skipping Gkfs Proxy [/][black][{get_time()}][/]"
        )
    else:
        console.print(
            f"[bold green]####### Starting Gkfs Proxy [/][black][{get_time()}][/]"
        )

        if settings.cluster:
            # Proxy call for cluster
            call = (
                f"srun --jobid={settings.job_id} {settings.app_nodes_command} "
                f"--disable-status -N {settings.app_nodes} --ntasks={settings.app_nodes} "
                f"--cpus-per-task={settings.procs_proxy} --ntasks-per-node=1 --overcommit "
                f"--overlap --oversubscribe --mem=0 {settings.task_set_1} {settings.gkfs_proxy} "
                f"-H {settings.gkfs_hostfile} -p ofi+verbs -P {settings.gkfs_proxyfile}"
            )
        else:
            # Proxy call for non-cluster
            call = (
                f"{settings.gkfs_proxy} -H {settings.gkfs_hostfile} "
                f"-p ofi+tcp -P {settings.gkfs_proxyfile}"
            )
        jit_print("[cyan]>> Starting Proxy[/]")

        # p = multiprocessing.Process(target=execute_background, args= (call, settings.gekko_proxy_log, settings.gekko_proxy_err, settings.dry_run))
        # p.start()
        # if settings.verbose:
        #     _ = monitor_log_file(settings.gekko_proxy_log,"Proxy")
        #     _ = monitor_log_file(settings.gekko_proxy_err,"Error Proxy")

        _ = execute_background_and_log(
            settings, call, settings.gekko_proxy_log, "proxy", settings.gekko_proxy_err
        )
        if settings.verbose:
            _ = monitor_log_file(settings.gekko_proxy_err, "Error Proxy")

        console.print("\n")


#! start Cargo
#!#################
def start_cargo(settings: JitSettings) -> None:
    if settings.exclude_cargo:
        console.print(
            f"[bold yellow]####### Skipping Cargo [/][black][{get_time()}][/]"
        )
    else:
        console.print(f"[bold green]####### Starting Cargo [/][black][{get_time()}][/]")

        if settings.cluster:
            if settings.use_mpirun:
                # mpiexec
                call = (
                    f"{settings.cargo} "
                    f"--listen {settings.gkfs_daemon_protocol}://ib0:62000 -b 65536"
                )
                call = flaged_mpiexec_call(
                    settings, call, settings.app_nodes * settings.procs_cargo
                )
            else:
                # srun
                call = (
                    f"srun --export=LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile},LIBGKFS_LOG_OUTPUT={settings.gekko_client_log}_cargo,LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile},"
                    f"LD_LIBRARY_PATH={os.environ.get('LD_LIBRARY_PATH')}{get_env(settings)} "
                    f"--jobid={settings.job_id} "
                    f"{settings.app_nodes_command} --disable-status "
                    f"-N {settings.app_nodes} --ntasks={settings.app_nodes*settings.procs_cargo} "
                    f"--cpus-per-task={settings.procs_cargo} --ntasks-per-node={settings.procs_cargo} "
                    f"--overcommit --overlap --oversubscribe --mem=0 {settings.cargo} "
                    f"--listen {settings.gkfs_daemon_protocol}://ib0:62000 -b 65536"
                )
        else:
            # Command for non-cluster
            call = f"{settings.cargo} --listen ofi+tcp://127.0.0.1:62000 -b 65536"
            call = flaged_mpiexec_call(settings, call, settings.procs_cargo)

        jit_print("[cyan]>> Starting Cargo[/]")

        # p_cargo = multiprocessing.Process(target=execute_background,args=(call, settings.cargo_log, settings.cargo_err, settings.dry_run))
        # p_cargo.start()
        # if settings.verbose:
        #     _ = monitor_log_file(settings.cargo_log,"Cargo")
        #     _ = monitor_log_file(settings.cargo_err,"Error Cargo")

        process = execute_background_and_log(
            settings, call, settings.cargo_log, "cargo", settings.cargo_err
        )
        if settings.verbose:
            _ = monitor_log_file(settings.cargo_err, "Error Cargo")
        # wait for line to appear
        time.sleep(2)
        flag = wait_for_line(
            settings.cargo_log, "Start up successful", dry_run=settings.dry_run
        )
        if not flag:
            handle_sigint(settings)
        time.sleep(4)
        console.print("\n")


#! start FTIO
#!################
def start_ftio(settings: JitSettings) -> None:
    if settings.exclude_ftio or settings.exclude_all:
        console.print(f"[bold yellow]####### Skipping FTIO [/][black][{get_time()}][/]")

        relevant_files(settings)

        if settings.cluster and not settings.exclude_cargo:
            console.print(
                "[bold yellow]Executing the calls below used later for staging out[/]"
            )
            call_0 = (
                f"srun --jobid={settings.job_id} {settings.single_node_command} "
                f"--disable-status -N 1 --ntasks=1 --cpus-per-task=1 --ntasks-per-node=1 "
                f"--overcommit --overlap --oversubscribe --mem=0 {settings.cargo_bin}/cargo_ftio "
                f"--server {settings.cargo_server} --run"
            )
            call_1 = (
                f"srun --jobid={settings.job_id} {settings.single_node_command} "
                f"--disable-status -N 1 --ntasks=1 --cpus-per-task=1 --ntasks-per-node=1 "
                f"--overcommit --overlap --oversubscribe --mem=0 {settings.cargo_bin}/ccp "
                f"--server {settings.cargo_server} --input / --output {settings.stage_out_path} "
                f"--if gekkofs --of {settings.cargo_mode}"
            )

            _ = execute_block(call_0, dry_run=settings.dry_run)
            _ = execute_block(call_1, dry_run=settings.dry_run)
    else:
        console.print(f"[bold green]####### Starting FTIO [/][black][{get_time()}][/]")

        relevant_files(settings)

        if settings.cluster:
            jit_print(
                f"[cyan]>> FTIO started on node {settings.ftio_node}, remaining nodes for the application: {settings.app_nodes} each with {settings.procs_ftio} processes [/]"
            )
            jit_print(
                f"[cyan]>> FTIO is listening node is {settings.address_ftio}:{settings.port} [/]"
            )

            call = (
                f"srun --jobid={settings.job_id} {settings.ftio_node_command} "
                f"--disable-status -N 1 --ntasks=1 --cpus-per-task={settings.procs_ftio} "
                f"--ntasks-per-node=1 --overcommit --overlap --oversubscribe --mem=0 "
                f"{settings.ftio_bin_location}/predictor_jit --zmq_address {settings.address_ftio} --zmq_port {settings.port} "
                f"--cargo_bin {settings.cargo_bin} --cargo_server {settings.cargo_server} "
                f"--cargo_out {settings.stage_out_path}"
            )
        else:
            check_port(settings)
            call = (
                f"{settings.ftio_bin_location}/predictor_jit --zmq_address {settings.address_ftio} --zmq_port {settings.port} "
                f"--cargo_bin {settings.cargo_bin} --cargo_server {settings.cargo_server} "
                f"--cargo_out {settings.stage_out_path}"
            )

        jit_print("[cyan]>> Starting FTIO[/]")

        process = execute_background_and_log(
            settings, call, settings.ftio_log, "ftio", settings.ftio_err
        )
        if settings.verbose:
            _ = monitor_log_file(settings.ftio_err, "Error Ftio")

        # p_ftio = multiprocessing.Process(target=execute_background,args=(call, settings.ftio_log, settings.ftio_err, settings.dry_run))
        # p_ftio.start()

        # if settings.verbose:
        #     _ = monitor_log_file(settings.ftio_log,"Ftio")
        #     _ = monitor_log_file(settings.ftio_err,"Error Ftio")

        if not settings.exclude_cargo:
            _ = wait_for_line(
                settings.ftio_log,
                "FTIO is running on:",
                "Waiting for FTIO startup",
                dry_run=settings.dry_run,
            )
        time.sleep(8)
        console.print("\n")


#! Stage in
#!################
def stage_in(settings: JitSettings, runtime: JitTime) -> None:
    if settings.exclude_all:
        console.print(
            f"[bold yellow]####### Skipping Stage in [/][black][{get_time()}][/]"
        )
    else:
        console.print(f"[bold green]####### Staging in [/][black][{get_time()}][/]")

        # remove locks
        jit_print("[cyan]>> Cleaning locks")
        if settings.stage_in_path:
            execute_block(
                f"cd {settings.stage_in_path} && rm -f  $(find .  | grep .lock)"
            )

        jit_print(f"[cyan]>> Staging in to {settings.stage_in_path}", True)

        if settings.exclude_cargo:
            # call = f"LD_PRELOAD={settings.gkfs_intercept} LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile} LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile} cp -r {settings.stage_in_path}/* {settings.gkfs_mntdir}"
            call = flaged_mpiexec_call(
                settings, f"cp -r {settings.stage_in_path}/* {settings.gkfs_mntdir}"
            )
            start = time.time()
            _ = execute_block(call, dry_run=settings.dry_run)
        else:
            if settings.cluster:
                call = f"srun --jobid={settings.job_id} {settings.single_node_command} --disable-status -N 1 --ntasks=1 --cpus-per-task=1 --ntasks-per-node=1 --overcommit --overlap --oversubscribe --mem=0 {settings.cargo_bin}/ccp --server {settings.cargo_server} --output / --input {settings.stage_in_path} --of gekkofs --if {settings.cargo_mode}"
            else:
                call = f"mpiexec -np 1 --oversubscribe {settings.cargo_bin}/ccp --server {settings.cargo_server} --output / --input {settings.stage_in_path} --of gekkofs --if {settings.cargo_mode}"

            start = time.time()

            execute_background_and_wait_line(
                call,
                settings.cargo_log,
                "retval: CARGO_SUCCESS, status: {state: completed",
                settings.dry_run,
            )

        elapsed_time(settings, runtime, "Stage in", time.time() - start)
        check(settings)


#! Stage out
#!################
def stage_out(settings: JitSettings, runtime: JitTime) -> None:
    if settings.exclude_all:
        console.print(
            f"[bold yellow]####### Skipping  Stage out [/][black][{get_time()}][/]"
        )
    else:
        if settings.exclude_cargo:
            # additional_arguments = load_flags(settings)
            # call = f"{additional_arguments} cp -r  {settings.gkfs_mntdir}/* {settings.stage_out_path} "
            call = flaged_mpiexec_call(
                settings,
                f"cp -r  {settings.gkfs_mntdir} {settings.stage_out_path} || echo 'nothing to stage in'",
            )
            start = time.time()
            _ = execute_block(call, dry_run=settings.dry_run)
            elapsed_time(settings, runtime, "Stage in", time.time() - start)
        else:
            console.print(
                f"[bold green]####### Staging out [/][black][{get_time()}][/]"
            )
            reset_relevant_files(settings)
            # time.sleep(8)

            if not settings.dry_run:
                try:
                    call = flaged_mpiexec_call(
                        settings, f"ls -R {settings.gkfs_mntdir}"
                    )
                    files = subprocess.check_output(call, shell=True).decode()
                    console.print(
                        f"[cyan]>> gekko_ls {settings.gkfs_mntdir}: \n{files}[/]"
                    )
                except Exception as e:
                    jit_print(f"[red] >> Error during test:\n{e}")

            # Reset relevant files
            if settings.cluster:
                call = (
                    f"srun --jobid={settings.job_id} {settings.single_node_command} "
                    f"--disable-status -N 1 --ntasks=1 --cpus-per-task=1 --ntasks-per-node=1 "
                    f"--overcommit --overlap --oversubscribe --mem=0 {settings.cargo_bin}/cargo_ftio "
                    f"--server {settings.cargo_server} --run"
                )
            else:
                call = (
                    f"mpiexec -np 1 --oversubscribe {settings.cargo_bin}/cargo_ftio "
                    f"--server {settings.cargo_server} --run"
                )

            # Measure and print elapsed time
            start = time.time()
            execute_background_and_wait_line(
                call,
                settings.cargo_log,
                # "[info] Transfer finished for [",
                "ftio_int RPC was successful!",
                settings.dry_run,
            )
            # end_of_transfer(settings, settings.cargo_log,call)
            end_of_transfer_online(settings, settings.cargo_log, call)
            elapsed_time(settings, runtime, "Stage out", time.time() - start)
            # Set ignored files to default again
            relevant_files(settings)
            time.sleep(5)


#! App call
#!################
def start_application(settings: JitSettings, runtime: JitTime):
    name = (
        settings.app_call.split("/", 1)[1]
        if "/" in settings.app_call
        else settings.app_call
    )
    console.print(
        f"[green bold]####### Executing Application {name} [/][black][{get_time()}][/]"
    )
    # set up dir
    original_dir = settings.dir
    jit_print(f">> Current directory {original_dir}")
    if not settings.dry_run:
        check_setup(settings)
        # pass

    if settings.cluster:
        additional_arguments = ""
        if settings.use_mpirun:
            if not settings.exclude_ftio:
                additional_arguments += f"-x LIBGKFS_METRICS_IP_PORT={settings.address_ftio}:{settings.port} -x LIBGKFS_ENABLE_METRICS=on "
            if not settings.exclude_proxy:
                additional_arguments += (
                    f"-x LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile} "
                )
            if not settings.exclude_daemon:
                additional_arguments += (
                    f"-x LIBGKFS_LOG=info,warnings,errors "
                    f"-x LIBGKFS_LOG_OUTPUT={settings.gekko_client_log} "
                    f"-x LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile} "
                    f"-x LD_PRELOAD={settings.gkfs_intercept} "
                )

            call = (
                # f" cd {settings.run_dir} && "
                # f"strace -f -e trace=read,write,open,close,stat,fstat,lseek,access -o /gpfs/fs1/home/tarrafah/strace_n{settings.app_nodes}_p{settings.procs_app}.txt mpiexec -np {settings.app_nodes*settings.procs_app} --oversubscribe "
                # f" cd {settings.run_dir} && time -p mpiexec --mca errhandler ftmpi --mca mpi_abort_print_stack 1  -np {settings.app_nodes*settings.procs_app} --oversubscribe "
                f" cd {settings.run_dir} && time -p mpiexec -np {settings.app_nodes*settings.procs_app} --oversubscribe "
                f"--hostfile {settings.dir}/hostfile_mpi --map-by node "
                f"{additional_arguments} "
                f"{settings.task_set_1} {settings.app_call} {settings.app_flags}"
            )
        else:
            if not settings.exclude_ftio:
                additional_arguments += f"LIBGKFS_ENABLE_METRICS=on,LIBGKFS_METRICS_IP_PORT={settings.address_ftio}:{settings.port},"
            if not settings.exclude_proxy:
                additional_arguments += (
                    f"LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile},"
                )
            if not settings.exclude_daemon:
                additional_arguments += (
                    f"LIBGKFS_LOG=info,warnings,errors,"
                    f"LIBGKFS_LOG_OUTPUT={settings.gekko_client_log},"
                    f"LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile},"
                    f"LD_PRELOAD={settings.gkfs_intercept},"
                )

            app_call = get_executable_realpath(settings.app_call,settings.run_dir)
            call = (
                f" cd {settings.run_dir} && time -p srun "
                f"--export=ALL,{additional_arguments}LD_LIBRARY_PATH={os.environ.get('LD_LIBRARY_PATH')} "
                f"--jobid={settings.job_id} {settings.app_nodes_command} --disable-status "
                f"-N {settings.app_nodes} --ntasks={settings.app_nodes*settings.procs_app} "
                f"--cpus-per-task={settings.procs_app} --ntasks-per-node={settings.procs_app} "
                f"--overcommit --overlap --oversubscribe --mem=0 "
                f"{app_call} {settings.app_flags}"
            )
    else:
        # Define the call for non-cluster environment
        if settings.run_dir:
            os.chdir(settings.run_dir)
            jit_print(f">> Changing directory to  {os.getcwd()}")

        if settings.exclude_all:
            call = f" time mpiexec -np {settings.procs_app} --oversubscribe {settings.app_call} {settings.app_flags}"
        elif settings.exclude_ftio:
            call = (
                f" time mpiexec -np {settings.procs_app} --oversubscribe "
                f"-x LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile} "
                f"-x LIBGKFS_LOG=info,warnings,errors -x LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile} "
                f"-x LD_PRELOAD={settings.gkfs_intercept} {settings.app_call} {settings.app_flags}"
            )
        else:
            call = (
                f" time mpiexec -np {settings.procs_app} --oversubscribe "
                f"-x LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile} "
                f"-x LIBGKFS_LOG=none -x LIBGKFS_ENABLE_METRICS=on "
                f"-x LIBGKFS_METRICS_IP_PORT={settings.address_ftio}:{settings.port} "
                f"-x LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile} "
                f"-x LD_PRELOAD={settings.gkfs_intercept} {settings.app_call} {settings.app_flags}"
            )

    # elapsed = execute_block_and_log(call, settings.app_log_dir)
    check(settings)
    start = time.time()
    # p_app = multiprocessing.Process(target=execute_background_and_log_in_process, args=(call, settings.app_log,"", settings.app_err, settings.dry_run))
    # p_app.start()
    # p_app.join()

    process = execute_background(
        call, settings.app_log, settings.app_err, settings.dry_run
    )
    if settings.verbose:
        _ = monitor_log_file(settings.app_log, "")
        _ = monitor_log_file(settings.app_err, f"{name} error")
    _, stderr = process.communicate()

    # get the real time
    # The timing result will be in stderr because 'time' outputs to stderr by default

    # Extract the 'real' time from the output
    real_time = time.time() - start
    try:
        with open(settings.app_err, "r") as file:
            for line in file:
                if line.startswith("real"):
                    real_time = min(float(line.split()[1]), real_time)
                    break
    except Exception as e:
        jit_print(f">>[red] Could not extract real time due to \n {e}[/]\n")

    elapsed_time(settings, runtime, "App", real_time)

    if process.returncode != 0:
        console.print(f"[red]Error executing command:{call}")
        console.print(f"[red] Error was:\n{stderr}")
    else:
        pass

    if not settings.exclude_ftio:
        jit_print("[cyan]>> Shuting down FTIO as application finished")
        shut_down(settings, "FTIO", settings.ftio_pid)

    os.chdir(original_dir)
    jit_print(f">> Changing directory to {os.getcwd()}")


#! Pre app call
#!################
def pre_call(settings: JitSettings) -> None:
    if settings.pre_app_call:
        console.print(
            f"[green bold]####### Pre-application Call [/][black][{get_time()}][/]"
        )
        # additional_arguments = load_flags(settings)
        if isinstance(settings.pre_app_call, str):
            call = settings.pre_app_call
            if any(x in call for x in ["mpiex", "mpirun"]):
                call = flaged_mpiexec_call(settings, call)
            execute_block_and_monitor(
                settings.verbose,
                call,
                settings.app_log,
                settings.app_err,
                settings.dry_run,
            )
        elif isinstance(settings.pre_app_call, list):
            for call in settings.pre_app_call:
                if any(x in call for x in ["mpiex", "mpirun"]):
                    call = flaged_mpiexec_call(settings, call)
                execute_block_and_monitor(
                    settings.verbose,
                    call,
                    settings.app_log,
                    settings.app_err,
                    settings.dry_run,
                )
        # _ = execute_block_and_log(
        #     settings.pre_app_call, settings.app_log
        # )


#! Post app call
#!################
def post_call(settings: JitSettings) -> None:
    if settings.post_app_call:
        console.print(
            f"[green bold]####### Post-application Call [/][black][{get_time()}][/]"
        )
        additional_arguments = ""  # load_flags(settings)
        if isinstance(settings.post_app_call, str):
            call = f"{additional_arguments} {settings.post_app_call}"
            execute_block_and_monitor(
                settings.verbose,
                call,
                settings.app_log,
                settings.app_err,
                settings.dry_run,
            )
        elif isinstance(settings.post_app_call, list):
            call = ""
            for s in settings.post_app_call:
                call = f"{additional_arguments} {s}"
                execute_block_and_monitor(
                    settings.verbose,
                    call,
                    settings.app_log,
                    settings.app_err,
                    settings.dry_run,
                )



import numpy as np
from sklearn.cluster import OPTICS, DBSCAN
import plotly.express as px
import pandas as pd

from ftio.plot.helper import format_plot


def optics(df:pd.DataFrame, s1:str, s2:str)-> pd.DataFrame:
    # Combine x and y into a 2D array
    data = np.column_stack((df[s1],df[s2]))

    # Apply OPTICS clustering
    optics = OPTICS(min_samples=5, xi=0.05, min_cluster_size=0.05)
    optics.fit(data)

    # Extract cluster labels
    labels = optics.labels_

    plot_cluster(df,s1,s2,labels)
    # plot_cluster_simple(df,s1,s2,labels)

    return  pd.concat([df, pd.DataFrame({"label":labels})], axis=1)


def dbscan(df: pd.DataFrame, s1: str, s2: str, eps = 0.5) -> pd.DataFrame:
    # Combine s1 and s2 columns into a 2D array for clustering
    data = np.column_stack((df[s1], df[s2]))

    # Apply DBSCAN clustering
    dbscan = DBSCAN(eps=eps, min_samples=5)
    dbscan.fit(data)

    # Extract cluster labels
    labels = dbscan.labels_

    plot_cluster(df, s1, s2, labels, "DBscan")

    return pd.concat([df, pd.DataFrame({"label": labels})], axis=1)


def plot_cluster_simple(df:pd.DataFrame, s1:str, s2:str, labels:np.ndarray, method="Optics") -> None:    
    # Plot the clusters
    fig = px.scatter(x=df[s1], y=df[s2], color=labels.astype(str), 
                    title=f"{method} Clustering",
                    labels={"color": "Cluster"},
                    color_discrete_sequence=px.colors.qualitative.Dark24)
    fig.show()

def plot_cluster(df: pd.DataFrame, s1: str, s2: str, labels: np.ndarray, method="Optics", unit_s1="rad", unit_s2="Hz") -> None:
    x = df[s1]
    y = df[s2]
    
    # Define unique labels and their corresponding colors and symbols
    labels = labels.astype(str)
    for i in range(len(labels)):
        if "-1" in labels[i]:
            labels[i] = "Outliers"
        else:
            labels[i] = f"Cluster {labels[i]}"
    
    unique_labels = set(labels)
    colors = px.colors.qualitative.Dark24[:len(unique_labels)]
    symbols = ["circle", "square", "diamond", "cross", "x", "triangle-up", "triangle-down"] * (len(unique_labels) // 7 + 1)
    
    color_map = {label: colors[i % len(colors)] for i, label in enumerate(unique_labels)}
    symbol_map = {label: symbols[i % len(symbols)] for i, label in enumerate(unique_labels)}

    # Map each point to its corresponding color and symbol
    point_colors = [color_map[label] for label in labels]
    point_symbols = [symbol_map[label] for label in labels]

    # Create a DataFrame for plotting
    df_plot = pd.DataFrame({
        f"{s1} ({unit_s1})": x,
        f"{s2} ({unit_s2})": y,
        "label": labels.astype(str),
        "color": point_colors,
        "symbol": point_symbols,
        "metric": df["Metric"]
    })

    # Plot the clusters using Plotly
    fig = px.scatter(data_frame=df_plot, 
                    x=f"{s1} ({unit_s1})", 
                    y=f"{s2} ({unit_s2})", 
                    title=f"{method} Clustering",
                    labels={"label":"Clusters"},
                    color="label",
                    color_discrete_map=color_map,
                    symbol="label",
                    symbol_map=symbol_map,
                    size_max=15)  # Set maximum size for markers

    # Customize the marker size and add hover information
    fig.update_traces(
        marker=dict(size=10),
        hovertemplate=(
            "x: %{x}<br>"
            "y: %{y}<br>"
            "Metric: %{customdata}<br>"
            "Label: %{marker.color}<br>"
            ))
    fig.update_layout(uniformtext_mode="hide")
    fig.update_traces(customdata=df["Metric"])
    fig = format_plot(fig)
    fig.show()


import os
import socket
import numpy as np
from rich.console import Console

console = Console()


class JitSettings:
    def __init__(self) -> None:
        """sets the internal variables, don't modify this part (except flags if needed).
        only Adjust the paths in the function set_variables
        """

        # flags
        ##############
        self.set_tasks_affinity = True  # required for ls and cp
        self.gkfs_daemon_protocol = (
            "ofi+verbs"  # "ofi+verbs" #"ofi+sockets"  or "ofi+verbs"
        )
        self.cargo_mode = "parallel"  # "parallel" or "posix"
        self.debug = True
        self.verbose = True
        self.node_local = False  # execute in node local space
        self.env_var = {}

        # Variable initialization (don't change)
        ################
        self.dry_run = False
        self.log_suffix = "DPCF"
        self.run_dir = ""
        self.dir = ""
        self.cluster = False
        self.job_id = 0
        self.static_allocation = False
        self.ftio_node = ""
        self.single_node = ""
        self.app_nodes = 0
        self.all_nodes = 0
        self.app_nodes_command = ""
        self.ftio_node_command = ""
        self.single_node_command = ""

        self.log_dir = ""
        self.gekko_daemon_log = ""
        self.gekko_daemon_err = ""
        self.gekko_proxy_log = ""
        self.gekko_proxy_err = ""
        self.gekko_client_log = ""
        self.cargo_log = ""
        self.cargo_err = ""
        self.ftio_log = ""
        self.ftio_err = ""
        self.app_log = ""
        self.app_err = ""

        # exclude flags
        self.exclude_ftio = False
        self.exclude_cargo = False
        self.exclude_daemon = False
        self.exclude_proxy = True
        self.exclude_all = False

        # pid of processes
        self.ftio_pid = 0
        self.gekko_daemon_pid = 0
        self.gekko_proxy_pid = 0
        self.cargo_pid = 0
        self.log_speed = 0.1

        # parsed variables
        ###################
        self.address_ftio = "127.0.0.1"
        self.address_cargo = "127.0.0.1"
        self.port = "5555"
        self.nodes = 1
        self.max_time = 30
        self.skip_confirm = False
        self.use_mpirun = False
        self.trap_exit = True
        self.soft_kill = True
        self.hard_kill = True

        self.procs = os.cpu_count() or 128
        self.omp_threads = 64
        self.task_set_0 = ""
        self.task_set_1 = ""
        self.procs_daemon = 0
        self.procs_proxy = 0
        self.procs_cargo = 0
        self.procs_app = 0
        self.procs_ftio = 0

        self.set_cluster_mode()
        self.set_default_procs()
        self.set_variables()

    def set_cluster_mode(self) -> None:
        """automatically identifies if it's a cluster or local machine"""
        hostname = socket.gethostname()
        if "cpu" in hostname or "mogon" in hostname:
            self.cluster = True
            if "mogon" in hostname:
                console.print(
                    "[bold red] Execute this script on CPU nodes\n mpiexec still has some bugs[/]"
                )

        console.print(f"[bold green]JIT >> CLUSTER MODE: {self.cluster}[/]")

    def update(self) -> None:
        """updates the flags and pass variables after the passed options are read.
        This is necessary, to adapt to the cluster mode and the installation path
        """
        self.set_flags()
        self.set_variables()
        self.set_absolute_path()
        self.update_settings()

    def update_settings(self):
        if self.dry_run:
            new_name = "Dry_" + self.job_name
            self.alloc_call_flags = self.alloc_call_flags.replace(
                self.job_name, new_name
            )
            self.job_name = new_name

    def set_absolute_path(self) -> None:
        self.run_dir = os.path.expanduser(self.run_dir)
        self.dir = os.path.expanduser(os.getcwd())
        self.ftio_bin_location = os.path.expanduser(self.ftio_bin_location)

    def set_default_procs(self) -> None:
        # default values for the procs in proc_list is not passed
        if self.cluster:
            self.procs_proxy = int(np.floor(self.procs / 2))
            self.procs_daemon = int(np.floor(self.procs / 2))
            self.procs_cargo = 2
            self.procs_ftio = self.procs
            self.procs_app = int(np.floor(self.procs / 2))
        else:
            self.procs = 10
            self.procs_daemon = 1
            self.procs_proxy = 1
            self.procs_cargo = 2
            self.procs_ftio = 1
            self.procs_app = self.procs

    def update_geko_files(self):
        if not self.exclude_daemon:
            self.gkfs_hostfile = self.gkfs_hostfile.replace(
                ".txt", f"_{self.job_id}.txt"
            )
        if not self.exclude_proxy:
            self.gkfs_proxyfile = self.gkfs_proxyfile.replace(
                ".pid", f"_{self.job_id}.pid"
            )

    def set_flags(self) -> None:
        """sets the flags in case exclude all is specified
        in the options passed
        """

        if (
            self.exclude_ftio
            and self.exclude_cargo
            and self.exclude_daemon
            and self.exclude_proxy
        ):
            self.exclude_all = True

        if self.exclude_all:
            self.exclude_ftio = True
            self.exclude_cargo = True
            self.exclude_daemon = True
            self.exclude_proxy = True

        if not self.cluster:
            if self.nodes > 1:
                self.procs = self.nodes
                self.nodes = 1
                console.print(
                    f"[bold green]JIT [bold cyan]>> correcting nodes to {self.nodes} and processes to {self.procs} [/]"
                )
        self.log_suffix = "DPCF"
        if self.exclude_daemon:
            self.procs_daemon = 0
            self.log_suffix = self.log_suffix.replace("D", "")
        if self.exclude_proxy:
            self.procs_proxy = 0
            self.log_suffix = self.log_suffix.replace("P", "")
        if self.exclude_cargo:
            self.procs_cargo = 0
            self.log_suffix = self.log_suffix.replace("C", "")
        if self.exclude_ftio:
            self.procs_ftio = 0
            self.log_suffix = self.log_suffix.replace("F", "")

        if self.set_tasks_affinity:
            self.task_set_0 = f"taskset -c 0-{np.floor(self.procs/2)-1:.0f}"
            if self.procs - np.floor(self.procs / 2) >= self.procs_app:
                self.task_set_1 = (
                    f"taskset -c {np.ceil(self.procs/2):.0f}-{self.procs-1:.0f}"
                )

    def set_log_dirs(self):
        self.gekko_daemon_log = os.path.join(self.log_dir, "gekko_daemon.log")
        self.gekko_daemon_err = os.path.join(self.log_dir, "gekko_daemon.err")
        self.gekko_proxy_log = os.path.join(self.log_dir, "gekko_proxy.log")
        self.gekko_proxy_err = os.path.join(self.log_dir, "gekko_proxy.err")
        self.gekko_client_log = os.path.join(self.log_dir, "gekko_client.log")
        self.cargo_log = os.path.join(self.log_dir, "cargo.log")
        self.cargo_err = os.path.join(self.log_dir, "cargo.err")
        self.ftio_log = os.path.join(self.log_dir, "ftio.log")
        self.ftio_err = os.path.join(self.log_dir, "ftio.err")
        self.app_log = os.path.join(self.log_dir, "app.log")
        self.app_err = os.path.join(self.log_dir, "app.err")

    def to_dict(self):  # -> dict[str, Any]:
        return {
            "nodes": self.nodes,
            "app nodes": self.app_nodes,
            "procs": self.procs,
            "procs daemon": self.procs_daemon,
            "procs proxy": self.procs_proxy,
            "procs cargo": self.procs_cargo,
            "procs app  ": self.procs_app,
            "procs ftio ": self.procs_ftio,
            "task_set_0": self.task_set_0,
            "task_set_1": self.task_set_1,
            "OpenMP threads": self.omp_threads,
            "protocol": self.gkfs_daemon_protocol,
            "run dir": self.run_dir,
            "app call": self.app_call,
            "id": self.job_id,
            "mode": self.log_suffix,
            "node local": self.node_local,
            "gkfs_mntdir": self.gkfs_mntdir if not self.exclude_daemon else "",
            "gkfs_rootdir": self.gkfs_rootdir if not self.exclude_daemon else "",
        }

    #!##########################
    #! only modify here
    #!##########################
    def set_variables(self) -> None:
        """sets the path variables"""
        # ****** install location ******
        if self.cluster:
            self.install_location = "/beegfs/home/Shared/admire/JIT"

        # ****** job allocation call ******
        # self.alloc_call_flags = "--overcommit --oversubscribe --partition parallel -A nhr-admire --job-name JIT --no-shell --exclude=cpu0082"
        self.job_name = "JIT"
        # self.alloc_call_flags = f"--overcommit --oversubscribe --partition largemem -A nhr-admire --job-name {self.job_name} --no-shell --exclude=cpu0081,cpu0082,cpu0083,cpu0084,cpu0401"
        self.alloc_call_flags = f"--overcommit --oversubscribe --partition parallel -A nhr-admire --job-name {self.job_name} --no-shell --exclude=cpu0081,cpu0082,cpu0083,cpu0084,cpu0401"

        # ? Tools
        # ?##########################
        # ****** ftio variables ******
        self.ftio_bin_location = "/lustre/project/nhr-admire/tarraf/FTIO/.venv/bin"

        # ****** gkfs variables ******
        self.gkfs_deps = "/lustre/project/nhr-admire/tarraf/deps"  # _gcc12_2"
        self.gkfs_daemon = f"{self.gkfs_deps}/gekkofs_zmq_install/bin/gkfs_daemon"
        self.gkfs_intercept = (
            # f"{self.gkfs_deps}/gekkofs_zmq_install/lib64/libgkfs_intercept.so" 
            f"{self.gkfs_deps}/gekkofs_zmq_install/lib64/libgkfs_libc_intercept.so" #from rnoun branch, doesn't work with nek
        )
        self.gkfs_mntdir = "/dev/shm/tarraf_gkfs_mountdir"
        self.gkfs_rootdir = "/dev/shm/tarraf_gkfs_rootdir"
        self.gkfs_hostfile = "/lustre/project/nhr-admire/tarraf/gkfs_hosts.txt"
        self.gkfs_proxy = (
            "/lustre/project/nhr-admire/tarraf/gekkofs/build/src/proxy/gkfs_proxy"
        )
        self.gkfs_proxyfile = "/dev/shm/tarraf_gkfs_proxy.pid"

        # ****** cargo variables ******
        self.cargo_bin = f"{self.gkfs_deps}/gekkofs_zmq_install/bin"  # "/lustre/project/nhr-admire/tarraf/cargo/build/cli"
        self.cargo = f"{self.cargo_bin}/cargo"  # "/lustre/project/nhr-admire/tarraf/cargo/build/src/cargo"
        self.cargo_server = f"{self.gkfs_daemon_protocol}://127.0.0.1:62000"

        # ? APP settings
        # ?##########################
        # ****** app call ******
        ##  ├─ IOR
        # self.app="/lustre/project/nhr-admire/tarraf/ior/src/ior -a POSIX -i 4 -o ${GKFS_MNTDIR}/iortest -t 128k -b 512m -F"
        # self.app="/lustre/project/nhr-admire/tarraf/HACC-IO/HACC_ASYNC_IO 1000000 ${GKFS_MNTDIR}/mpi"
        ##  ├─ DLIO -->
        self.app_call = "dlio_benchmark"
        self.run_dir = "."
        self.app_flags = (
            f"workload=unet3d_my_a100 "
            # f"workload=bert_v100 "
            f"++workload.workflow.generate_data=True ++workload.workflow.train=True ++workload.workflow.checkpoint=True " #++workload.workflow.evaluation=True "
            f"++workload.dataset.data_folder={self.run_dir}/data/jit ++workload.checkpoint.checkpoint_folder={self.run_dir}/checkpoints/jit "  
            f"++workload.output.output_folder={self.run_dir}/hydra_log/jit "
        )
        ##  ├─ LAMMPS -->
        # self.app_call = "/lustre/project/nhr-admire/shared/mylammps/build/lmp"
        # self.run_dir = f"{self.gkfs_mntdir}"
        # self.app_flags = "-in in.spce.hex"
        # ##  ├─ NEK5000 --> change gkfs_daemon_protocol to socket
        # self.app_call = "./nek5000"
        # self.run_dir = "/home/tarrafah/nhr-admire/shared/run_gkfs_marc"
        # self.app_flags = ""
        ##  └─ Wacom++ --> change wacom.json if needed
        # self.app_call = "./wacommplusplus"
        # self.run_dir = "/lustre/project/nhr-admire/tarraf/wacommplusplus/build"#_gcc12_2"
        # self.app_flags = ""

        # ****** pre and post app call ******
        # Application specific calls executed before the actual run. Executed as
        # > ${PRE_APP_CALL}
        # > cd self.run_dir && mpiexec ${some flags} ..${APP_CALL}
        # > ${POST_APP_CALL}
        # ├─ dlio
        if "dlio" in self.app_call:
            if self.exclude_all:
                # self.pre_app_call = f"mpirun -np 8 dlio_benchmark {self.app_flags} ++workload.workflow.generate_data=True ++workload.workflow.train=False"
                self.pre_app_call  = ""
                self.post_app_call = ""
            else:
                # self.run_dir = self.gkfs_mntdir #? don't enable this flag, as the executing node doesn't have this folder
                self.app_flags = (
                    # f"workload=unet3d_my_a100_gekko "
                    f"workload=unet3d_my_a100 " #no need to select a specific one, the folder assignment makes it use geko
                    # f"workload=bert_v100 "
                    f"++workload.workflow.generate_data=True ++workload.workflow.train=True ++workload.workflow.checkpoint=True " #++workload.workflow.evaluation=True "
                    f"++workload.dataset.data_folder={self.gkfs_mntdir}/data/jit ++workload.checkpoint.checkpoint_folder={self.gkfs_mntdir}/checkpoints/jit " 
                    f"++workload.output.output_folder={self.gkfs_mntdir}/hydra_log/jit "
                    # ++workload.output.log_file={self.gkfs_mntdir}/hydra_log/unet3d"# ++workload.dataset.num_files_train=16"
                )
                # self.app_flags = "workload=unet3d_my_a100_gekko"
                # self.pre_app_call = f"mpirun -np 10 dlio_benchmark {self.app_flags} ++workload.workflow.generate_data=True ++workload.workflow.train=False"
                self.pre_app_call  = ""
                self.post_app_call = ""
        # ├─ Nek5000
        elif "nek" in self.app_call:
            if self.exclude_all:
                self.pre_app_call = f"echo -e 'turbPipe\\n{self.run_dir}/input' > {self.run_dir}/SESSION.NAME"
                self.post_app_call = f"rm {self.run_dir}/input/*.f* || echo true"
            else:
                self.pre_app_call = f"echo -e 'turbPipe\\n{self.gkfs_mntdir}' > {self.run_dir}/SESSION.NAME"
                self.post_app_call = ""
        # ├─ Wacom++
        elif "wacom" in self.app_call:
            if self.exclude_all:
                # in case a previous simulation fails
                self.pre_app_call = (
                    f"export OMP_NUM_THREADS={self.omp_threads}; ln -sf {self.run_dir}/wacomm.pfs.json {self.run_dir}/wacomm.json; "
                    f"cd {self.run_dir} && rm -rf input restart results processed output; cp -r stage-in/* . "
                )
                self.post_app_call = ""
            else:
                # modify wacomm.gkfs.json to include gkfs_mntdir
                self.pre_app_call = f"export OMP_NUM_THREADS={self.omp_threads}; ln -sf {self.run_dir}/wacomm.gkfs.json {self.run_dir}/wacomm.json; "
                self.post_app_call = (
                    f"ln -sf {self.run_dir}/wacomm.pfs.json {self.run_dir}/wacomm.json"
                )
        # └─ Other
        else:
            self.pre_app_call = ""
            self.post_app_call = ""

        # ? Stage in/out
        # ?##########################
        # ├─ Nek5000
        if "nek" in self.app_call:
            self.stage_in_path = f"{self.run_dir}/input"
            self.stage_out_path = "/lustre/project/nhr-admire/tarraf/stage-out"
        # ├─ Wacom++
        elif "wacom" in self.app_call:
            self.stage_in_path = f"{self.run_dir}/stage-in"
            self.stage_out_path = "/lustre/project/nhr-admire/tarraf/stage-out"
        # ├─ Wacom++
        elif "lmp" in self.app_call:
            self.stage_in_path = "/lustre/project/nhr-admire/shared/mylammps/examples/HEAT"
            self.stage_out_path = "/lustre/project/nhr-admire/tarraf/stage-out"
        # └─ Other
        else:
            self.stage_in_path =  "/lustre/project/nhr-admire/tarraf/stage-in"
            self.stage_out_path = "/lustre/project/nhr-admire/tarraf/stage-out"

        # ? Regex if needed
        # ?##########################
        self.regex_file = "/lustre/project/nhr-admire/shared/nek_regex4cargo.txt"
        # ├─ Nek5000
        if "nek" in self.app_call:
            self.regex_match = "^/[a-zA-Z0-9]*turbPipe0\\.f\\d+"
        # ├─ Wacom++
        elif "wacom" in self.app_call:
            # self.regex_match = "^(\\/(output|results|restart|input|processed)\\/)\\.+$"  #
            self.regex_match = "\\/processed\\/^ocm3_d03_\\d+Z\\d+\\.nc$"
        # ├─ LAMMPS
        elif "lmp" in self.app_call:
            self.regex_match = ""
        # └─ Other
        else:
            self.regex_match = ""

        self.env_var = {"CARGO_REGEX": self.regex_file}

        # ? local machine settings
        # ?###############################
        if not self.cluster:
            self.install_location = "/d/github/JIT"
            self.ftio_bin_location = "/d/github/FTIO/.venv/bin"
            self.gkfs_daemon = f"{self.install_location}/iodeps/bin/gkfs_daemon"
            self.gkfs_intercept = (
                # f"{self.install_location}/iodeps/lib/libgkfs_intercept.so"
                f"{self.install_location}/iodeps/lib/libgkfs_libc_intercept.so"
            )
            self.gkfs_mntdir = "/tmp/jit/tarraf_gkfs_mountdir"
            self.gkfs_rootdir = "/tmp/jit/tarraf_gkfs_rootdir"
            self.gkfs_hostfile = f"{os.getcwd()}/gkfs_hosts.txt"
            self.gkfs_proxy = (
                f"{self.install_location}/gekkofs/build/src/proxy/gkfs_proxy"
            )
            self.gkfs_proxyfile = f"{self.install_location}/tarraf_gkfs_proxy.pid"
            self.cargo = f"{self.install_location}/cargo/build/src/cargo"
            self.cargo_bin = f"{self.install_location}/cargo/build/cli"

            self.regex_file = "/tmp/jit/nek_regex4cargo.txt"
            self.env_var = {"CARGO_REGEX": self.regex_file}

            if "dlio" in self.app_call:
                self.stage_in_path = "/d/github/dlio_benchmark/data"
                # generate data with
                if self.exclude_all:
                    self.run_dir = ""  # don't change dir
                    self.app_flags = "workload=unet3d_my_a100"
                    self.pre_app_call = f"mpirun -np 8 dlio_benchmark {self.app_flags} ++workload.workflow.generate_data=True ++workload.workflow.train=False"
                    self.post_app_call = ""
                else:
                    self.run_dir = f"{self.gkfs_mntdir}"
                    self.app_flags = "workload=unet3d_my_a100_gekko"
                    self.pre_app_call = [
                        # f"cd {self.gkfs_mntdir}",
                        f"mpirun -np 8 dlio_benchmark {self.app_flags} ++workload.workflow.generate_data=True ++workload.workflow.train=False",
                        # with one proc it works
                    ]
                    self.post_app_call = ""
                # ├─ Nek5000
            elif "nek" in self.app_call:
                self.run_dir = "/d/benchmark/Nek5000/turbPipe/run"
                self.stage_in_path = "/d/benchmark/Nek5000/turbPipe/run/input"
                if self.exclude_all:
                    self.pre_app_call = "echo -e 'turbPipe\\n/d/benchmark/Nek5000/turbPipe/run/input' > /d/benchmark/Nek5000/turbPipe/run/SESSION.NAME"
                    self.post_app_call = (
                        "rm /d/benchmark/Nek5000/turbPipe/run/input/*.f* || true"
                    )
                else:
                    self.pre_app_call = f"echo -e 'turbPipe\\n{self.gkfs_mntdir}' > /d/benchmark/Nek5000/turbPipe/run/SESSION.NAME"
                    self.post_app_call = f"rm {self.stage_out_path}/*.f* || true"

            self.stage_out_path = "/tmp/output"
            # self.stage_in_path = "/tmp/input"
            # self.stage_out_path = "/tmp/output"



from ftio.api.gekkoFs.jit.jitsettings import JitSettings
from ftio.api.gekkoFs.jit.setup_helper import jit_print, create_hostfile
from ftio.api.gekkoFs.jit.execute_and_wait import execute_block

def init_gekko(settings: JitSettings) -> None:
    """Creates GekkoFs hostfile and directories (root and mount)

    Args:
        settings (JitSettings): Jit settings
    """
    if not settings.exclude_daemon:
        create_hostfile(settings)
        calls =[]
        # set debug flag
        if settings.cluster:
            # Create directories
            # calls_0 = f"srun --jobid={settings.job_id} {settings.single_node_command} -N 1 --ntasks=1 mkdir -p {settings.gkfs_mntdir}"
            calls.append(
                f"srun --jobid={settings.job_id} {settings.app_nodes_command} --disable-status -N {settings.app_nodes} "
                f"--ntasks={settings.app_nodes} --cpus-per-task={settings.procs_daemon} --ntasks-per-node=1 --overcommit --overlap "
                f"--oversubscribe --mem=0 mkdir -p {settings.gkfs_mntdir}"
                    )
            calls.append(
                f"srun --jobid={settings.job_id} {settings.app_nodes_command} --disable-status -N {settings.app_nodes} "
                f"--ntasks={settings.app_nodes} --cpus-per-task={settings.procs_daemon} --ntasks-per-node=1 --overcommit --overlap "
                f"--oversubscribe --mem=0 mkdir -p {settings.gkfs_rootdir}"
                    )
        else:
            calls.append(f"mkdir -p {settings.gkfs_mntdir}")
            calls.append(f"mkdir -p {settings.gkfs_rootdir}")

        jit_print("[cyan]>> Creating directories[/]")
        for call in calls:
            _ = execute_block(call)

